#ifndef lint
static char sccs_id[] = "%W%	(TriChlor) %G% %U%";
#endif	/* lint */

#include "portable.h"
#include "vntype.h"
#include "charset.h"
#include "vnkeys.h"

static u_char _8to7map[256][4]; /* at most <a><^><'><NUL> */

static _doinit() { /* fill the map--the code is independent of the encoding */
    /* C0 ASCII section---must be first to be overridden by Vietnamese */
    _8to7map[C_SOH][0] = C_SOH;
    _8to7map[C_STX][0] = C_STX;
    _8to7map[C_ETX][0] = C_ETX;
    _8to7map[C_EOT][0] = C_EOT;
    _8to7map[C_ENQ][0] = C_ENQ;
    _8to7map[C_ACK][0] = C_ACK;
    _8to7map[C_BEL][0] = C_BEL;
    _8to7map[C_BS][0] = C_BS;
    _8to7map[C_HT][0] = C_HT;
    _8to7map[C_LF][0] = C_LF;
    _8to7map[C_VT][0] = C_VT;
    _8to7map[C_FF][0] = C_FF;
    _8to7map[C_CR][0] = C_CR;
    _8to7map[C_SO][0] = C_SO;
    _8to7map[C_SI][0] = C_SI;
    _8to7map[C_DLE][0] = C_DLE;
    _8to7map[C_DC1][0] = C_DC1;
    _8to7map[C_DC2][0] = C_DC2;
    _8to7map[C_DC3][0] = C_DC3;
    _8to7map[C_DC4][0] = C_DC4;
    _8to7map[C_NAK][0] = C_NAK;
    _8to7map[C_SYN][0] = C_SYN;
    _8to7map[C_ETB][0] = C_ETB;
    _8to7map[C_CAN][0] = C_CAN;
    _8to7map[C_EM][0] = C_EM;
    _8to7map[C_SUB][0] = C_SUB;
    _8to7map[C_ESC][0] = C_ESC;
    _8to7map[C_IS4][0] = C_IS4;
    _8to7map[C_IS3][0] = C_IS3;
    _8to7map[C_IS2][0] = C_IS2;
    _8to7map[C_IS1][0] = C_IS1;

    /* C0 Vietnamese section */
    _8to7map[C_A_breve_hook_above][0] = C_A;
    _8to7map[C_A_breve_hook_above][1] = C_breve;
    _8to7map[C_A_breve_hook_above][2] = C_hook_above;
    _8to7map[C_A_breve_tilde][0] = C_A;
    _8to7map[C_A_breve_tilde][1] = C_breve;
    _8to7map[C_A_breve_tilde][2] = C_tilde;
    _8to7map[C_A_circumflex_tilde][0] = C_A;
    _8to7map[C_A_circumflex_tilde][1] = C_caret;
    _8to7map[C_A_circumflex_tilde][2] = C_tilde;
    _8to7map[C_Y_hook_above][0] = C_Y;
    _8to7map[C_Y_hook_above][1] = C_hook_above;
    _8to7map[C_y_hook_above/C_IS3][0] = C_y;
    _8to7map[C_y_hook_above/C_IS3][1] = C_hook_above;
    _8to7map[C_Y_tilde][0] = C_Y;
    _8to7map[C_Y_tilde][1] = C_tilde;
    _8to7map[C_Y_dot_below][0] = C_Y;
    _8to7map[C_Y_dot_below][1] = C_dot_below;

    _8to7map[C_space][0] = C_space;
    _8to7map[C_exclamation_mark][0] = C_exclamation_mark;
    _8to7map[C_quotation_mark][0] = C_quotation_mark;
    _8to7map[C_pound_sign][0] = C_pound_sign;
    _8to7map[C_dollar_sign][0] = C_dollar_sign;
    _8to7map[C_percent][0] = C_percent;
    _8to7map[C_ampersand][0] = C_ampersand;
    _8to7map[C_apostrophe][0] = C_apostrophe;
    _8to7map[C_left_parenthesis][0] = C_left_parenthesis;
    _8to7map[C_right_parenthesis][0] = C_right_parenthesis;
    _8to7map[C_asterisk][0] = C_asterisk;
    _8to7map[C_plus_sign][0] = C_plus_sign;
    _8to7map[C_comma][0] = C_comma;
    _8to7map[C_hyphen][0] = C_hyphen;
    _8to7map[C_period][0] = C_period;
    _8to7map[C_slash][0] = C_slash;

    _8to7map[C_zero][0] = C_zero;
    _8to7map[C_one][0] = C_one;
    _8to7map[C_two][0] = C_two;
    _8to7map[C_three][0] = C_three;
    _8to7map[C_four][0] = C_four;
    _8to7map[C_five][0] = C_five;
    _8to7map[C_six][0] = C_six;
    _8to7map[C_seven][0] = C_seven;
    _8to7map[C_eight][0] = C_eight;
    _8to7map[C_nine][0] = C_nine;

    _8to7map[C_colon][0] = C_colon;
    _8to7map[C_semicolon][0] = C_semicolon;
    _8to7map[C_less_than][0] = C_less_than;
    _8to7map[C_equals_sign][0] = C_equals_sign;
    _8to7map[C_greater_than][0] = C_greater_than;
    _8to7map[C_hook_above][0] = C_hook_above;
    _8to7map[C_commercial_at][0] = C_commercial_at;

    _8to7map[C_A][0] = C_A;
    _8to7map[C_B][0] = C_B;
    _8to7map[C_C][0] = C_C;
    _8to7map[C_D][0] = C_D;
    _8to7map[C_E][0] = C_E;
    _8to7map[C_F][0] = C_F;
    _8to7map[C_G][0] = C_G;
    _8to7map[C_H][0] = C_H;
    _8to7map[C_I][0] = C_I;
    _8to7map[C_J][0] = C_J;
    _8to7map[C_K][0] = C_K;
    _8to7map[C_L][0] = C_L;
    _8to7map[C_M][0] = C_M;
    _8to7map[C_N][0] = C_N;
    _8to7map[C_O][0] = C_O;
    _8to7map[C_P][0] = C_P;
    _8to7map[C_Q][0] = C_Q;
    _8to7map[C_R][0] = C_R;
    _8to7map[C_S][0] = C_S;
    _8to7map[C_T][0] = C_T;
    _8to7map[C_U][0] = C_U;
    _8to7map[C_V][0] = C_V;
    _8to7map[C_W][0] = C_W;
    _8to7map[C_X][0] = C_X;
    _8to7map[C_Y][0] = C_Y;
    _8to7map[C_Z][0] = C_Z;

    _8to7map[C_left_square_bracket][0] = C_left_square_bracket;
    _8to7map[C_back_slash][0] = C_back_slash;
    _8to7map[C_right_square_bracket][0] = C_right_square_bracket;
    _8to7map[C_caret][0] = C_caret;
    _8to7map[C_underscore][0] = C_underscore;
    _8to7map[C_backquote][0] = C_backquote;

    _8to7map[C_a][0] = C_a;
    _8to7map[C_b][0] = C_b;
    _8to7map[C_c][0] = C_c;
    _8to7map[C_d][0] = C_d;
    _8to7map[C_e][0] = C_e;
    _8to7map[C_f][0] = C_f;
    _8to7map[C_g][0] = C_g;
    _8to7map[C_h][0] = C_h;
    _8to7map[C_i][0] = C_i;
    _8to7map[C_j][0] = C_j;
    _8to7map[C_k][0] = C_k;
    _8to7map[C_l][0] = C_l;
    _8to7map[C_m][0] = C_m;
    _8to7map[C_n][0] = C_n;
    _8to7map[C_o][0] = C_o;
    _8to7map[C_p][0] = C_p;
    _8to7map[C_q][0] = C_q;
    _8to7map[C_r][0] = C_r;
    _8to7map[C_s][0] = C_s;
    _8to7map[C_t][0] = C_t;
    _8to7map[C_u][0] = C_u;
    _8to7map[C_v][0] = C_v;
    _8to7map[C_w][0] = C_w;
    _8to7map[C_x][0] = C_x;
    _8to7map[C_y][0] = C_y;
    _8to7map[C_z][0] = C_z;

    _8to7map[C_left_brace][0] = C_left_brace;
    _8to7map[C_vertical_bar][0] = C_vertical_bar;
    _8to7map[C_right_brace][0] = C_right_brace;
    _8to7map[C_tilde][0] = C_tilde;
    _8to7map[C_delete][0] = C_delete;

    /* C1 & G1 Vietnamese section */
    _8to7map[C_A_hook_above][0] = C_A;
    _8to7map[C_A_hook_above][1] = C_hook_above;
    _8to7map[C_A_dot_below][0] = C_A;
    _8to7map[C_A_dot_below][1] = C_dot_below;
    _8to7map[C_A_breve][0] = C_A;
    _8to7map[C_A_breve][1] = C_breve;
    _8to7map[C_A_breve_acute][0] = C_A;
    _8to7map[C_A_breve_acute][1] = C_breve; C_acute;
    _8to7map[C_A_breve_acute][2] = C_acute;
    _8to7map[C_A_breve_grave][0] = C_A;
    _8to7map[C_A_breve_grave][1] = C_breve;
    _8to7map[C_A_breve_grave][2] = C_grave;
    _8to7map[C_A_breve_dot_below][0] = C_A;
    _8to7map[C_A_breve_dot_below][1] = C_breve;
    _8to7map[C_A_breve_dot_below][2] = C_dot_below;
    _8to7map[C_A_circumflex_acute][0] = C_A;
    _8to7map[C_A_circumflex_acute][1] = C_circumflex;
    _8to7map[C_A_circumflex_acute][2] = C_acute;
    _8to7map[C_A_circumflex_grave][0] = C_A;
    _8to7map[C_A_circumflex_grave][1] = C_circumflex;
    _8to7map[C_A_circumflex_grave][2] = C_grave;
    _8to7map[C_A_circumflex_hook_above][0] = C_A;
    _8to7map[C_A_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_A_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_A_circumflex_dot_below][0] = C_A;
    _8to7map[C_A_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_A_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_E_hook_above][0] = C_E;
    _8to7map[C_E_hook_above][1] = C_hook_above;
    _8to7map[C_E_tilde][0] = C_E;
    _8to7map[C_E_tilde][1] = C_tilde;
    _8to7map[C_E_dot_below][0] = C_E;
    _8to7map[C_E_dot_below][1] = C_dot_below;
    _8to7map[C_E_circumflex_acute][0] = C_E;
    _8to7map[C_E_circumflex_acute][1] = C_circumflex;
    _8to7map[C_E_circumflex_acute][2] = C_acute;
    _8to7map[C_E_circumflex_grave][0] = C_E;
    _8to7map[C_E_circumflex_grave][1] = C_circumflex;
    _8to7map[C_E_circumflex_grave][2] = C_grave;
    _8to7map[C_E_circumflex_hook_above][0] = C_E;
    _8to7map[C_E_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_E_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_E_circumflex_tilde][0] = C_E;
    _8to7map[C_E_circumflex_tilde][1] = C_circumflex;
    _8to7map[C_E_circumflex_tilde][2] = C_tilde;
    _8to7map[C_E_circumflex_dot_below][0] = C_E;
    _8to7map[C_E_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_E_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_I_hook_above][0] = C_I;
    _8to7map[C_I_hook_above][1] = C_hook_above;
    _8to7map[C_I_tilde][0] = C_I;
    _8to7map[C_I_tilde][1] = C_tilde;
    _8to7map[C_I_dot_below][0] = C_I;
    _8to7map[C_I_dot_below][1] = C_dot_below;
    _8to7map[C_O_hook_above][0] = C_O;
    _8to7map[C_O_hook_above][1] = C_hook_above;
    _8to7map[C_O_dot_below][0] = C_O;
    _8to7map[C_O_dot_below][1] = C_dot_below;
    _8to7map[C_O_circumflex_acute][0] = C_O;
    _8to7map[C_O_circumflex_acute][1] = C_circumflex;
    _8to7map[C_O_circumflex_acute][2] = C_acute;
    _8to7map[C_O_circumflex_grave][0] = C_O;
    _8to7map[C_O_circumflex_grave][1] = C_circumflex;
    _8to7map[C_O_circumflex_grave][2] = C_grave;
    _8to7map[C_O_circumflex_hook_above][0] = C_O;
    _8to7map[C_O_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_O_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_O_circumflex_tilde][0] = C_O;
    _8to7map[C_O_circumflex_tilde][1] = C_circumflex;
    _8to7map[C_O_circumflex_tilde][2] = C_tilde;
    _8to7map[C_O_circumflex_dot_below][0] = C_O;
    _8to7map[C_O_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_O_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_O_horn][0] = C_O;
    _8to7map[C_O_horn][1] = C_horn;
    _8to7map[C_O_horn_acute][0] = C_O;
    _8to7map[C_O_horn_acute][1] = C_horn;
    _8to7map[C_O_horn_acute][2] = C_acute;
    _8to7map[C_O_horn_grave][0] = C_O;
    _8to7map[C_O_horn_grave][1] = C_horn;
    _8to7map[C_O_horn_grave][2] = C_grave;
    _8to7map[C_O_horn_hook_above][0] = C_O;
    _8to7map[C_O_horn_hook_above][1] = C_horn;
    _8to7map[C_O_horn_hook_above][2] = C_hook_above;

    _8to7map[C_a_hook_above][0] = C_a;
    _8to7map[C_a_hook_above][1] = C_hook_above;
    _8to7map[C_a_dot_below][0] = C_a;
    _8to7map[C_a_dot_below][1] = C_dot_below;
    _8to7map[C_a_breve][0] = C_a;
    _8to7map[C_a_breve][1] = C_breve;
    _8to7map[C_a_breve_acute][0] = C_a;
    _8to7map[C_a_breve_acute][1] = C_breve;
    _8to7map[C_a_breve_acute][2] = C_acute;
    _8to7map[C_a_breve_grave][0] = C_a;
    _8to7map[C_a_breve_grave][1] = C_breve;
    _8to7map[C_a_breve_grave][2] = C_grave;
    _8to7map[C_a_breve_dot_below][0] = C_a;
    _8to7map[C_a_breve_dot_below][1] = C_breve;
    _8to7map[C_a_breve_dot_below][2] = C_dot_below;
    _8to7map[C_a_circumflex_acute][0] = C_a;
    _8to7map[C_a_circumflex_acute][1] = C_circumflex;
    _8to7map[C_a_circumflex_acute][2] = C_acute;
    _8to7map[C_a_circumflex_grave][0] = C_a;
    _8to7map[C_a_circumflex_grave][1] = C_circumflex;
    _8to7map[C_a_circumflex_grave][2] = C_grave;
    _8to7map[C_a_circumflex_hook_above][0] = C_a;
    _8to7map[C_a_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_a_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_a_circumflex_dot_below][0] = C_a;
    _8to7map[C_a_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_a_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_e_hook_above][0] = C_e;
    _8to7map[C_e_hook_above][1] = C_hook_above;
    _8to7map[C_e_tilde][0] = C_e;
    _8to7map[C_e_tilde][1] = C_tilde;
    _8to7map[C_e_dot_below][0] = C_e;
    _8to7map[C_e_dot_below][1] = C_dot_below;
    _8to7map[C_e_circumflex_acute][0] = C_e;
    _8to7map[C_e_circumflex_acute][1] = C_circumflex;
    _8to7map[C_e_circumflex_acute][2] = C_acute;
    _8to7map[C_e_circumflex_grave][0] = C_e;
    _8to7map[C_e_circumflex_grave][1] = C_circumflex;
    _8to7map[C_e_circumflex_grave][2] = C_grave;
    _8to7map[C_e_circumflex_hook_above][0] = C_e;
    _8to7map[C_e_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_e_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_e_circumflex_tilde][0] = C_e;
    _8to7map[C_e_circumflex_tilde][1] = C_circumflex;
    _8to7map[C_e_circumflex_tilde][2] = C_tilde;
    _8to7map[C_e_circumflex_dot_below][0] = C_e;
    _8to7map[C_e_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_e_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_i_hook_above][0] = C_i;
    _8to7map[C_i_hook_above][1] = C_hook_above;
    _8to7map[C_U_horn_tilde][0] = C_U;
    _8to7map[C_U_horn_tilde][1] = C_horn;
    _8to7map[C_U_horn_tilde][2] = C_tilde;
    _8to7map[C_U_horn_dot_below][0] = C_U;
    _8to7map[C_U_horn_dot_below][1] = C_horn;
    _8to7map[C_U_horn_dot_below][2] = C_dot_below;
    _8to7map[C_o_hook_above][0] = C_o;
    _8to7map[C_o_hook_above][1] = C_hook_above;
    _8to7map[C_o_dot_below][0] = C_o;
    _8to7map[C_o_dot_below][1] = C_dot_below;
    _8to7map[C_o_circumflex_acute][0] = C_o;
    _8to7map[C_o_circumflex_acute][1] = C_circumflex;
    _8to7map[C_o_circumflex_acute][2] = C_acute;
    _8to7map[C_o_circumflex_grave][0] = C_o;
    _8to7map[C_o_circumflex_grave][1] = C_circumflex;
    _8to7map[C_o_circumflex_grave][2] = C_grave;
    _8to7map[C_o_circumflex_hook_above][0] = C_o;
    _8to7map[C_o_circumflex_hook_above][1] = C_circumflex;
    _8to7map[C_o_circumflex_hook_above][2] = C_hook_above;
    _8to7map[C_o_circumflex_tilde][0] = C_o;
    _8to7map[C_o_circumflex_tilde][1] = C_circumflex;
    _8to7map[C_o_circumflex_tilde][2] = C_tilde;
    _8to7map[C_o_circumflex_dot_below][0] = C_o;
    _8to7map[C_o_circumflex_dot_below][1] = C_circumflex;
    _8to7map[C_o_circumflex_dot_below][2] = C_dot_below;
    _8to7map[C_o_horn][0] = C_o;
    _8to7map[C_o_horn][1] = C_horn;
    _8to7map[C_o_horn_acute][0] = C_o;
    _8to7map[C_o_horn_acute][1] = C_horn;
    _8to7map[C_o_horn_acute][2] = C_acute;
    _8to7map[C_o_horn_grave][0] = C_o;
    _8to7map[C_o_horn_grave][1] = C_horn;
    _8to7map[C_o_horn_grave][2] = C_grave;
    _8to7map[C_U_horn_hook_above][0] = C_U;
    _8to7map[C_U_horn_hook_above][1] = C_horn;
    _8to7map[C_U_horn_hook_above][2] = C_hook_above;
    _8to7map[C_A_grave][0] = C_A;
    _8to7map[C_A_grave][1] = C_grave;
    _8to7map[C_A_acute][0] = C_A;
    _8to7map[C_A_acute][1] = C_acute;
    _8to7map[C_A_circumflex][0] = C_A;
    _8to7map[C_A_circumflex][1] = C_circumflex;
    _8to7map[C_A_tilde][0] = C_A;
    _8to7map[C_A_tilde][1] = C_tilde;
    _8to7map[C_O_horn_tilde][0] = C_O;
    _8to7map[C_O_horn_tilde][1] = C_horn;
    _8to7map[C_O_horn_tilde][2] = C_tilde;
    _8to7map[C_O_horn_dot_below][0] = C_O;
    _8to7map[C_O_horn_dot_below][1] = C_horn;
    _8to7map[C_O_horn_dot_below][2] = C_dot_below;
    _8to7map[C_U_hook_above][0] = C_U;
    _8to7map[C_U_hook_above][1] = C_hook_above;
    _8to7map[C_U_tilde][0] = C_U;
    _8to7map[C_U_tilde][1] = C_tilde;
    _8to7map[C_E_grave][0] = C_E;
    _8to7map[C_E_grave][1] = C_grave;
    _8to7map[C_E_acute][0] = C_E;
    _8to7map[C_E_acute][1] = C_acute;
    _8to7map[C_E_circumflex][0] = C_E;
    _8to7map[C_E_circumflex][1] = C_circumflex;
    _8to7map[C_U_dot_below][0] = C_U;
    _8to7map[C_U_dot_below][1] = C_dot_below;
    _8to7map[C_I_grave][0] = C_I;
    _8to7map[C_I_grave][1] = C_grave;
    _8to7map[C_I_acute][0] = C_I;
    _8to7map[C_I_acute][1] = C_acute;
    _8to7map[C_U_horn][0] = C_U;
    _8to7map[C_U_horn][1] = C_horn;
    _8to7map[C_U_horn_acute][0] = C_U;
    _8to7map[C_U_horn_acute][1] = C_horn;
    _8to7map[C_U_horn_acute][2] = C_acute;
    _8to7map[C_D_bar][0] = C_D;
    _8to7map[C_D_bar][1] = C_D;
    _8to7map[C_U_horn_grave][0] = C_U;
    _8to7map[C_U_horn_grave][1] = C_horn;
    _8to7map[C_U_horn_grave][2] = C_grave;
    _8to7map[C_O_grave][0] = C_O;
    _8to7map[C_O_grave][1] = C_grave;
    _8to7map[C_O_acute][0] = C_O;
    _8to7map[C_O_acute][1] = C_acute;
    _8to7map[C_O_circumflex][0] = C_O;
    _8to7map[C_O_circumflex][1] = C_circumflex;
    _8to7map[C_O_tilde][0] = C_O;
    _8to7map[C_O_tilde][1] = C_tilde;
    _8to7map[C_o_horn_hook_above][0] = C_o;
    _8to7map[C_o_horn_hook_above][1] = C_horn;
    _8to7map[C_o_horn_hook_above][2] = C_hook_above;
    _8to7map[C_i_tilde][0] = C_i;
    _8to7map[C_i_tilde][1] = C_tilde;
    _8to7map[C_i_dot_below][0] = C_i;
    _8to7map[C_i_dot_below][1] = C_dot_below;
    _8to7map[C_U_grave][0] = C_U;
    _8to7map[C_U_grave][1] = C_grave;
    _8to7map[C_U_acute][0] = C_U;
    _8to7map[C_U_acute][1] = C_acute;
    _8to7map[C_Y_grave][0] = C_Y;
    _8to7map[C_Y_grave][1] = C_grave;
    _8to7map[C_y_hook_above][0] = C_y;
    _8to7map[C_y_hook_above][1] = C_hook_above;
    _8to7map[C_Y_acute][0] = C_Y;
    _8to7map[C_Y_acute][1] = C_acute;
    _8to7map[C_y_tilde][0] = C_y;
    _8to7map[C_y_tilde][1] = C_tilde;
    _8to7map[C_y_dot_below][0] = C_y;
    _8to7map[C_y_dot_below][1] = C_dot_below;
    _8to7map[C_a_grave][0] = C_a;
    _8to7map[C_a_grave][1] = C_grave;
    _8to7map[C_a_acute][0] = C_a;
    _8to7map[C_a_acute][1] = C_acute;
    _8to7map[C_a_circumflex][0] = C_a;
    _8to7map[C_a_circumflex][1] = C_circumflex;
    _8to7map[C_a_tilde][0] = C_a;
    _8to7map[C_a_tilde][1] = C_tilde;
    _8to7map[C_o_horn_tilde][0] = C_o;
    _8to7map[C_o_horn_tilde][1] = C_horn;
    _8to7map[C_o_horn_tilde][2] = C_tilde;
    _8to7map[C_o_horn_dot_below][0] = C_o;
    _8to7map[C_o_horn_dot_below][1] = C_horn;
    _8to7map[C_o_horn_dot_below][2] = C_dot_below;
    _8to7map[C_u_hook_above][0] = C_u;
    _8to7map[C_u_hook_above][1] = C_hook_above;
    _8to7map[C_u_tilde][0] = C_u;
    _8to7map[C_u_tilde][1] = C_tilde;
    _8to7map[C_e_grave][0] = C_e;
    _8to7map[C_e_grave][1] = C_grave;
    _8to7map[C_e_acute][0] = C_e;
    _8to7map[C_e_acute][1] = C_acute;
    _8to7map[C_e_circumflex][0] = C_e;
    _8to7map[C_e_circumflex][1] = C_circumflex;
    _8to7map[C_u_dot_below][0] = C_u;
    _8to7map[C_u_dot_below][1] = C_dot_below;
    _8to7map[C_i_grave][0] = C_i;
    _8to7map[C_i_grave][1] = C_grave;
    _8to7map[C_i_acute][0] = C_i;
    _8to7map[C_i_acute][1] = C_acute;
    _8to7map[C_u_horn][0] = C_u;
    _8to7map[C_u_horn][1] = C_horn;
    _8to7map[C_u_horn_acute][0] = C_u;
    _8to7map[C_u_horn_acute][1] = C_horn;
    _8to7map[C_u_horn_acute][2] = C_acute;
    _8to7map[C_d_bar][0] = C_d;
    _8to7map[C_d_bar][1] = C_d;
    _8to7map[C_u_horn_grave][0] = C_u;
    _8to7map[C_u_horn_grave][1] = C_horn;
    _8to7map[C_u_horn_grave][2] = C_grave;
    _8to7map[C_o_grave][0] = C_o;
    _8to7map[C_o_grave][1] = C_grave;
    _8to7map[C_o_acute][0] = C_o;
    _8to7map[C_o_acute][1] = C_acute;
    _8to7map[C_o_circumflex][0] = C_o;
    _8to7map[C_o_circumflex][1] = C_circumflex;
    _8to7map[C_o_tilde][0] = C_o;
    _8to7map[C_o_tilde][1] = C_tilde;
    _8to7map[C_u_horn_hook_above][0] = C_u;
    _8to7map[C_u_horn_hook_above][1] = C_horn;
    _8to7map[C_u_horn_hook_above][2] = C_hook_above;
    _8to7map[C_u_horn_tilde][0] = C_u;
    _8to7map[C_u_horn_tilde][1] = C_horn;
    _8to7map[C_u_horn_tilde][2] = C_tilde;
    _8to7map[C_u_horn_dot_below][0] = C_u;
    _8to7map[C_u_horn_dot_below][1] = C_horn;
    _8to7map[C_u_horn_dot_below][2] = C_dot_below;
    _8to7map[C_u_grave][0] = C_u;
    _8to7map[C_u_grave][1] = C_grave;
    _8to7map[C_u_acute][0] = C_u;
    _8to7map[C_u_acute][1] = C_acute;
    _8to7map[C_y_grave][0] = C_y;
    _8to7map[C_y_grave][1] = C_grave;
    _8to7map[C_a_breve_hook_above][0] = C_a;
    _8to7map[C_a_breve_hook_above][1] = C_breve;
    _8to7map[C_a_breve_hook_above][2] = C_hook_above;
    _8to7map[C_y_acute][0] = C_y;
    _8to7map[C_y_acute][0] = C_y;
    _8to7map[C_y_acute][1] = C_acute;
    _8to7map[C_a_breve_tilde][0] = C_a;
    _8to7map[C_a_breve_tilde][1] = C_breve;
    _8to7map[C_a_breve_tilde][2] = C_tilde;
    _8to7map[C_a_circumflex_tilde][0] = C_a;
    _8to7map[C_a_circumflex_tilde][1] = C_circumflex;
    _8to7map[C_a_circumflex_tilde][2] = C_tilde;
}


/*
 * vk_8to7_vietnamese: Vietnamese vk_8to7().  See comments for vk_8to7().
 *		       Generates an "esc" for the following:
 *			- the esc itself
 *			- between two d's (upper- or lower-case)
 *			- between a vowel and a diacritic
 */
static int
vk_8to7_vietnamese(inbuf, outbuf, esc)
    u_char	*inbuf;
    u_char	*outbuf;
    u_char	esc;
{
    register u_char	c;
    register u_char	prevc;
    register u_char	*i, *o, *s;

    prevc = '\0';
    i = inbuf;
    o = outbuf;

    /* Try to arrange the tests from fastest to slowest */
    while (c = *i++) {
	if ( (c == esc) ||
	     (tolower(c) == 'd' && tolower(prevc) == 'd') ||
	     (ismodifier(c) && isvowel(prevc) && !ismodified(prevc)) ||
	     (isaccent(c) && isvowel(prevc) && !isaccented(prevc))) {
	    *o++ = esc;
	    *o++ = c;
	} else {
	    s = _8to7map[c];
	    while (*o = *s++) o++;
	}
	prevc = c;
    }

    *o = '\0';
    return((int) (o - outbuf));
}


/*
 * vk_8to7_english: English vk_8to7().	See comments for vk_8to7().
 *		    Generates an esc for the following:
 *			- the esc itself
 *			- an accented/modified vowel
 *			- the Vietnamese letter d-bar
 */
static int
vk_8to7_english(inbuf, outbuf, esc)
    u_char	*inbuf;
    u_char	*outbuf;
    u_char	esc;
{
    register u_char	c;
    register u_char	prevc;
    register u_char	*i, *o, *s;

    prevc = '\0';
    i = inbuf;
    o = outbuf;

    /* Try to arrange the tests from fastest to slowest */
    while (c = *i++) {
	if ( (c == esc) || isdiacriticized(c) || (tolower(c) == C_d_bar) ) {
	    *o++ = esc;
	}
	s = _8to7map[c];
	while (*o = *s++) o++;
	prevc = c;
    }

    *o = '\0';
    return((int) (o - outbuf));
}


/*
 * vk_8to7: converts the 8-bit Vietnamese "inbuf" into 7-bit
 *	    vnkeys-compliant data, and stores the result in
 *	    "outbuf".  The escape character is passed in through "esc".
 *	    The bias (Vietnamese or English) is specified with "state".
 *	    Both inbuf and outbuf must be allocated and supplied
 *	    by the caller.  Inbuf is assumed to be null-terminated.
 *
 *	    Returns the size of the actual outbuf data, in bytes,
 *	    not including the terminating NULL.
 *	    The returning outbuf is null-terminated.
 *
 *	    Note that vk_8to7() does NOT generate a leading
 *	    \V or \M in the returned outbuf.
 *
 *	    The conversion to LITERAL is NOT supported because it
 *	    won't see too much use.
 */
int FAR PASCAL
#ifdef ANSI_C
vk_8toqr(u_char FAR *inbuf, u_char FAR *outbuf, u_char esc, Vk_State state)
#else
vk_8toqr(inbuf, outbuf, esc, state)
    u_char FAR *inbuf;
    u_char FAR *outbuf;
    u_char esc;
    Vk_State state;
#endif
{
    static int	inited = 0;

    if (!inited) {
        _doinit();	/* private initialization function */
	inited = 1;
    }

    switch(state) {
    case VK_ST_ENGLISH:
	return vk_8to7_english(inbuf, outbuf, esc);
    case VK_ST_VIETNAMESE:
	return vk_8to7_vietnamese(inbuf, outbuf, esc);
    default:
	return 0;
    }
}
